@@ -11,26 +11,31 @@ import logging
 import uuid
 
 from datetime import datetime
-from signal import (
-        signal, default_int_handler, SIGINT
-)
+try:
+    # jupyter_client >= 5, use tz-aware now
+    from jupyter_client.session import utcnow as now
+except ImportError:
+    # jupyter_client < 5, use local now()
+    now = datetime.now
+
+from signal import signal, default_int_handler, SIGINT
 
 import zmq
-from zmq.eventloop import ioloop
+from tornado import ioloop
 from zmq.eventloop.zmqstream import ZMQStream
 
 from traitlets.config.configurable import SingletonConfigurable
 from IPython.core.error import StdinNotImplementedError
-from IPython.core import release
 from ipython_genutils import py3compat
 from ipython_genutils.py3compat import unicode_type, string_types
 from ipykernel.jsonutil import json_clean
 from traitlets import (
-    Any, Instance, Float, Dict, List, Set, Integer, Unicode, Bool,
+    Any, Instance, Float, Dict, List, Set, Integer, Unicode, Bool, observe, default
 )
 
 from jupyter_client.session import Session
 
+from ._version import kernel_protocol_version
 
 class Kernel(SingletonConfigurable):
 
@@ -40,24 +45,29 @@ class Kernel(SingletonConfigurable):
 
     # attribute to override with a GUI
     eventloop = Any(None)
-    def _eventloop_changed(self, name, old, new):
+
+    @observe('eventloop')
+    def _update_eventloop(self, change):
         """schedule call to eventloop from IOLoop"""
-        loop = ioloop.IOLoop.instance()
-        loop.add_callback(self.enter_eventloop)
+        loop = ioloop.IOLoop.current()
+        if change.new is not None:
+            loop.add_callback(self.enter_eventloop)
 
     session = Instance(Session, allow_none=True)
     profile_dir = Instance('IPython.core.profiledir.ProfileDir', allow_none=True)
     shell_streams = List()
     control_stream = Instance(ZMQStream, allow_none=True)
-    iopub_socket = Instance(zmq.Socket, allow_none=True)
-    stdin_socket = Instance(zmq.Socket, allow_none=True)
+    iopub_socket = Any()
+    iopub_thread = Any()
+    stdin_socket = Any()
     log = Instance(logging.Logger, allow_none=True)
 
     # identities:
     int_id = Integer(-1)
     ident = Unicode()
 
-    def _ident_default(self):
+    @default('ident')
+    def _default_ident(self):
         return unicode_type(uuid.uuid4())
 
     # This should be overridden by wrapper kernels that implement any real
@@ -69,12 +79,12 @@ class Kernel(SingletonConfigurable):
 
     # Private interface
 
-    _darwin_app_nap = Bool(True, config=True,
+    _darwin_app_nap = Bool(True,
         help="""Whether to use appnope for compatiblity with OS X App Nap.
 
         Only affects OS X >= 10.9.
         """
-    )
+    ).tag(config=True)
 
     # track associations with current request
     _allow_stdin = Bool(False)
@@ -87,12 +97,12 @@ class Kernel(SingletonConfigurable):
     # Units are in seconds.  The minimum zmq latency on local host is probably
     # ~150 microseconds, set this to 500us for now.  We may need to increase it
     # a little if it's not enough after more interactive testing.
-    _execute_sleep = Float(0.0005, config=True)
+    _execute_sleep = Float(0.0005).tag(config=True)
 
     # Frequency of the kernel's event loop.
     # Units are in seconds, kernel subclasses for GUI toolkits may need to
     # adapt to milliseconds.
-    _poll_interval = Float(0.05, config=True)
+    _poll_interval = Float(0.05).tag(config=True)
 
     # If the shutdown was requested over the network, we leave here the
     # necessary reply message so it can be sent by our registered atexit
@@ -112,27 +122,29 @@ class Kernel(SingletonConfigurable):
     # execution count we store in the shell.
     execution_count = 0
 
+    msg_types = [
+        'execute_request', 'complete_request',
+        'inspect_request', 'history_request',
+        'comm_info_request', 'kernel_info_request',
+        'connect_request', 'shutdown_request',
+        'is_complete_request',
+        # deprecated:
+        'apply_request',
+    ]
+    # add deprecated ipyparallel control messages
+    control_msg_types = msg_types + ['clear_request', 'abort_request']
 
     def __init__(self, **kwargs):
         super(Kernel, self).__init__(**kwargs)
-
         # Build dict of handlers for message types
-        msg_types = [ 'execute_request', 'complete_request',
-                      'inspect_request', 'history_request',
-                      'kernel_info_request',
-                      'connect_request', 'shutdown_request',
-                      'apply_request', 'is_complete_request',
-                    ]
         self.shell_handlers = {}
-        for msg_type in msg_types:
+        for msg_type in self.msg_types:
             self.shell_handlers[msg_type] = getattr(self, msg_type)
 
-        control_msg_types = msg_types + [ 'clear_request', 'abort_request' ]
         self.control_handlers = {}
-        for msg_type in control_msg_types:
+        for msg_type in self.control_msg_types:
             self.control_handlers[msg_type] = getattr(self, msg_type)
 
-
     def dispatch_control(self, msg):
         """dispatch control requests"""
         idents,msg = self.session.feed_identities(msg, copy=False)
@@ -164,6 +176,25 @@ class Kernel(SingletonConfigurable):
         sys.stderr.flush()
         self._publish_status(u'idle')
 
+    def should_handle(self, stream, msg, idents):
+        """Check whether a shell-channel message should be handled
+
+        Allows subclasses to prevent handling of certain messages (e.g. aborted requests).
+        """
+        msg_id = msg['header']['msg_id']
+        if msg_id in self.aborted:
+            msg_type = msg['header']['msg_type']
+            # is it safe to assume a msg_id will not be resubmitted?
+            self.aborted.remove(msg_id)
+            reply_type = msg_type.split('_')[0] + '_reply'
+            status = {'status' : 'aborted'}
+            md = {'engine' : self.ident}
+            md.update(status)
+            self.session.send(stream, reply_type, metadata=md,
+                        content=status, parent=msg, ident=idents)
+            return False
+        return True
+
     def dispatch_shell(self, stream, msg):
         """dispatch shell requests"""
         # flush control requests first
@@ -181,8 +212,6 @@ class Kernel(SingletonConfigurable):
         self.set_parent(idents, msg)
         self._publish_status(u'busy')
 
-        header = msg['header']
-        msg_id = header['msg_id']
         msg_type = msg['header']['msg_type']
 
         # Print some info about this message and leave a '--->' marker, so it's
@@ -191,35 +220,35 @@ class Kernel(SingletonConfigurable):
         self.log.debug('\n*** MESSAGE TYPE:%s***', msg_type)
         self.log.debug('   Content: %s\n   --->\n   ', msg['content'])
 
-        if msg_id in self.aborted:
-            self.aborted.remove(msg_id)
-            # is it safe to assume a msg_id will not be resubmitted?
-            reply_type = msg_type.split('_')[0] + '_reply'
-            status = {'status' : 'aborted'}
-            md = {'engine' : self.ident}
-            md.update(status)
-            self.session.send(stream, reply_type, metadata=md,
-                        content=status, parent=msg, ident=idents)
+        if not self.should_handle(stream, msg, idents):
             return
 
         handler = self.shell_handlers.get(msg_type, None)
         if handler is None:
-            self.log.error("UNKNOWN MESSAGE TYPE: %r", msg_type)
+            self.log.warning("Unknown message type: %r", msg_type)
         else:
-            # ensure default_int_handler during handler call
-            sig = signal(SIGINT, default_int_handler)
             self.log.debug("%s: %s", msg_type, msg)
+            self.pre_handler_hook()
             try:
                 handler(stream, idents, msg)
             except Exception:
                 self.log.error("Exception in message handler:", exc_info=True)
             finally:
-                signal(SIGINT, sig)
+                self.post_handler_hook()
 
         sys.stdout.flush()
         sys.stderr.flush()
         self._publish_status(u'idle')
 
+    def pre_handler_hook(self):
+        """Hook to execute before calling message handler"""
+        # ensure default_int_handler during handler call
+        self.saved_sigint_handler = signal(SIGINT, default_int_handler)
+
+    def post_handler_hook(self):
+        """Hook to execute after calling message handler"""
+        signal(SIGINT, self.saved_sigint_handler)
+
     def enter_eventloop(self):
         """enter eventloop"""
         self.log.info("entering eventloop %s", self.eventloop)
@@ -228,7 +257,7 @@ class Kernel(SingletonConfigurable):
             # which may be skipped by entering the eventloop
             stream.flush(zmq.POLLOUT)
         # restore default_int_handler
-        signal(SIGINT, default_int_handler)
+        self.pre_handler_hook()
         while self.eventloop is not None:
             try:
                 self.eventloop(self)
@@ -240,10 +269,12 @@ class Kernel(SingletonConfigurable):
                 # eventloop exited cleanly, this means we should stop (right?)
                 self.eventloop = None
                 break
+        self.post_handler_hook()
         self.log.info("exiting eventloop")
 
     def start(self):
         """register dispatchers for streams"""
+        self.io_loop = ioloop.IOLoop.current()
         if self.control_stream:
             self.control_stream.on_recv(self.dispatch_control, copy=False)
 
@@ -267,7 +298,6 @@ class Kernel(SingletonConfigurable):
             stream.flush(zmq.POLLIN, 1)
             stream.flush(zmq.POLLOUT)
 
-
     def record_ports(self, ports):
         """Record the ports that this kernel is using.
 
@@ -280,17 +310,6 @@ class Kernel(SingletonConfigurable):
     # Kernel request handlers
     #---------------------------------------------------------------------------
 
-    def _make_metadata(self, other=None):
-        """init metadata dict, for execute/apply_reply"""
-        new_md = {
-            'dependencies_met' : True,
-            'engine' : self.ident,
-            'started': datetime.now(),
-        }
-        if other:
-            new_md.update(other)
-        return new_md
-
     def _publish_execute_input(self, code, parent, execution_count):
         """Publish the code request on the iopub stream."""
 
@@ -333,6 +352,24 @@ class Kernel(SingletonConfigurable):
         return self.session.send(stream, msg_or_type, content, self._parent_header,
                                  ident, buffers, track, header, metadata)
 
+    def init_metadata(self, parent):
+        """Initialize metadata.
+
+        Run at the beginning of execution requests.
+        """
+        # FIXME: `started` is part of ipyparallel
+        # Remove for ipykernel 5.0
+        return {
+            'started': now(),
+        }
+
+    def finish_metadata(self, parent, metadata, reply_content):
+        """Finish populating metadata.
+
+        Run after completing an execution request.
+        """
+        return metadata
+
     def execute_request(self, stream, ident, parent):
         """handle an execute_request"""
 
@@ -350,7 +387,7 @@ class Kernel(SingletonConfigurable):
 
         stop_on_error = content.get('stop_on_error', True)
 
-        md = self._make_metadata(parent['metadata'])
+        metadata = self.init_metadata(parent)
 
         # Re-broadcast our input for the benefit of listening clients, and
         # start computing output
@@ -372,14 +409,10 @@ class Kernel(SingletonConfigurable):
 
         # Send the reply.
         reply_content = json_clean(reply_content)
-
-        md['status'] = reply_content['status']
-        if reply_content['status'] == 'error' and \
-                        reply_content['ename'] == 'UnmetDependency':
-                md['dependencies_met'] = False
+        metadata = self.finish_metadata(parent, metadata, reply_content)
 
         reply_msg = self.session.send(stream, u'execute_reply',
-                                      reply_content, parent, metadata=md,
+                                      reply_content, parent, metadata=metadata,
                                       ident=ident)
 
         self.log.debug("%s", reply_msg)
@@ -397,12 +430,11 @@ class Kernel(SingletonConfigurable):
         content = parent['content']
         code = content['code']
         cursor_pos = content['cursor_pos']
-
+        
         matches = self.do_complete(code, cursor_pos)
         matches = json_clean(matches)
         completion_msg = self.session.send(stream, 'complete_reply',
                                            matches, parent, ident)
-        self.log.debug("%s", completion_msg)
 
     def do_complete(self, code, cursor_pos):
         """Override in subclasses to find completions.
@@ -427,7 +459,7 @@ class Kernel(SingletonConfigurable):
     def do_inspect(self, code, cursor_pos, detail_level=0):
         """Override in subclasses to allow introspection.
         """
-        return {'status': 'ok', 'data':{}, 'metadata':{}, 'found':False}
+        return {'status': 'ok', 'data': {}, 'metadata': {}, 'found': False}
 
     def history_request(self, stream, ident, parent):
         content = parent['content']
@@ -443,13 +475,14 @@ class Kernel(SingletonConfigurable):
                    stop=None, n=None, pattern=None, unique=False):
         """Override in subclasses to access history.
         """
-        return {'history': []}
+        return {'status': 'ok', 'history': []}
 
     def connect_request(self, stream, ident, parent):
         if self._recorded_ports is not None:
             content = self._recorded_ports.copy()
         else:
             content = {}
+        content['status'] = 'ok'
         msg = self.session.send(stream, 'connect_reply',
                                 content, parent, ident)
         self.log.debug("%s", msg)
@@ -457,7 +490,7 @@ class Kernel(SingletonConfigurable):
     @property
     def kernel_info(self):
         return {
-            'protocol_version': release.kernel_protocol_version,
+            'protocol_version': kernel_protocol_version,
             'implementation': self.implementation,
             'implementation_version': self.implementation_version,
             'language_info': self.language_info,
@@ -466,8 +499,28 @@ class Kernel(SingletonConfigurable):
         }
 
     def kernel_info_request(self, stream, ident, parent):
+        content = {'status': 'ok'}
+        content.update(self.kernel_info)
         msg = self.session.send(stream, 'kernel_info_reply',
-                                self.kernel_info, parent, ident)
+                                content, parent, ident)
+        self.log.debug("%s", msg)
+
+    def comm_info_request(self, stream, ident, parent):
+        content = parent['content']
+        target_name = content.get('target_name', None)
+
+        # Should this be moved to ipkernel?
+        if hasattr(self, 'comm_manager'):
+            comms = {
+                k: dict(target_name=v.target_name)
+                for (k, v) in self.comm_manager.comms.items()
+                if v.target_name == target_name or target_name is None
+            }
+        else:
+            comms = {}
+        reply_content = dict(comms=comms, status='ok')
+        msg = self.session.send(stream, 'comm_info_reply',
+                                reply_content, parent, ident)
         self.log.debug("%s", msg)
 
     def shutdown_request(self, stream, ident, parent):
@@ -480,7 +533,7 @@ class Kernel(SingletonConfigurable):
 
         self._at_shutdown()
         # call sys.exit after a short delay
-        loop = ioloop.IOLoop.instance()
+        loop = ioloop.IOLoop.current()
         loop.add_timeout(time.time()+0.1, loop.stop)
 
     def do_shutdown(self, restart):
@@ -506,10 +559,11 @@ class Kernel(SingletonConfigurable):
                 }
 
     #---------------------------------------------------------------------------
-    # Engine methods
+    # Engine methods (DEPRECATED)
     #---------------------------------------------------------------------------
 
     def apply_request(self, stream, ident, parent):
+        self.log.warning("apply_request is deprecated in kernel_base, moving to ipyparallel.")
         try:
             content = parent[u'content']
             bufs = parent[u'buffers']
@@ -518,31 +572,30 @@ class Kernel(SingletonConfigurable):
             self.log.error("Got bad msg: %s", parent, exc_info=True)
             return
 
-        md = self._make_metadata(parent['metadata'])
+        md = self.init_metadata(parent)
 
         reply_content, result_buf = self.do_apply(content, bufs, msg_id, md)
 
-        # put 'ok'/'error' status in header, for scheduler introspection:
-        md['status'] = reply_content['status']
-
         # flush i/o
         sys.stdout.flush()
         sys.stderr.flush()
 
+        md = self.finish_metadata(parent, md, reply_content)
+
         self.session.send(stream, u'apply_reply', reply_content,
                     parent=parent, ident=ident,buffers=result_buf, metadata=md)
 
     def do_apply(self, content, bufs, msg_id, reply_metadata):
-        """Override in subclasses to support the IPython parallel framework.
-        """
+        """DEPRECATED"""
         raise NotImplementedError
 
     #---------------------------------------------------------------------------
-    # Control messages
+    # Control messages (DEPRECATED)
     #---------------------------------------------------------------------------
 
     def abort_request(self, stream, ident, parent):
         """abort a specific msg by id"""
+        self.log.warning("abort_request is deprecated in kernel_base. It os only part of IPython parallel")
         msg_ids = parent['content'].get('msg_ids', None)
         if isinstance(msg_ids, string_types):
             msg_ids = [msg_ids]
@@ -558,15 +611,13 @@ class Kernel(SingletonConfigurable):
 
     def clear_request(self, stream, idents, parent):
         """Clear our namespace."""
+        self.log.warning("clear_request is deprecated in kernel_base. It os only part of IPython parallel")
         content = self.do_clear()
         self.session.send(stream, 'clear_reply', ident=idents, parent=parent,
                 content = content)
 
     def do_clear(self):
-        """Override in subclasses to clear the namespace
-
-        This is only required for IPython.parallel.
-        """
+        """DEPRECATED"""
         raise NotImplementedError
 
     #---------------------------------------------------------------------------
@@ -575,10 +626,7 @@ class Kernel(SingletonConfigurable):
 
     def _topic(self, topic):
         """prefixed topic for IOPub messages"""
-        if self.int_id >= 0:
-            base = "engine.%i" % self.int_id
-        else:
-            base = "kernel.%s" % self.ident
+        base = "kernel.%s" % self.ident
 
         return py3compat.cast_bytes("%s.%s" % (base, topic))
 
@@ -603,21 +651,22 @@ class Kernel(SingletonConfigurable):
             status = {'status' : 'aborted'}
             md = {'engine' : self.ident}
             md.update(status)
+            self._publish_status('busy', parent=msg)
             reply_msg = self.session.send(stream, reply_type, metadata=md,
                         content=status, parent=msg, ident=idents)
+            self._publish_status('idle', parent=msg)
             self.log.debug("%s", reply_msg)
             # We need to wait a bit for requests to come in. This can probably
             # be set shorter for true asynchronous clients.
             poller.poll(50)
 
-
     def _no_raw_input(self):
         """Raise StdinNotImplentedError if active frontend doesn't support
         stdin."""
         raise StdinNotImplementedError("raw_input was called, but this "
                                        "frontend does not support stdin.")
 
-    def getpass(self, prompt=''):
+    def getpass(self, prompt='', stream=None):
         """Forward getpass to frontends
 
         Raises
@@ -628,6 +677,10 @@ class Kernel(SingletonConfigurable):
             raise StdinNotImplementedError(
                 "getpass was called, but this frontend does not support input requests."
             )
+        if stream is not None:
+            import warnings
+            warnings.warn("The `stream` parameter of `getpass.getpass` will have no effect when using ipykernel",
+                    UserWarning, stacklevel=2)
         return self._input_request(prompt,
             self._parent_ident,
             self._parent_header,
@@ -645,7 +698,7 @@ class Kernel(SingletonConfigurable):
             raise StdinNotImplementedError(
                 "raw_input was called, but this frontend does not support input requests."
             )
-        return self._input_request(prompt,
+        return self._input_request(str(prompt),
             self._parent_ident,
             self._parent_header,
             password=False,
@@ -675,7 +728,7 @@ class Kernel(SingletonConfigurable):
             try:
                 ident, reply = self.session.recv(self.stdin_socket, 0)
             except Exception:
-                self.log.warn("Invalid Message:", exc_info=True)
+                self.log.warning("Invalid Message:", exc_info=True)
             except KeyboardInterrupt:
                 # re-raise KeyboardInterrupt, to truncate traceback
                 raise KeyboardInterrupt
@@ -694,7 +747,7 @@ class Kernel(SingletonConfigurable):
     def _at_shutdown(self):
         """Actions taken at shutdown by the kernel, called by python's atexit.
         """
-        # io.rprint("Kernel at_shutdown") # dbg
+        # io.raw_print("Kernel at_shutdown") # dbg
         if self._shutdown_message is not None:
             self.session.send(self.iopub_socket, self._shutdown_message, ident=self._topic('shutdown'))
             self.log.debug("%s", self._shutdown_message)
